#!/usr/bin/env python

"""
Fundamental program data structure abstractions.

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# Short representation display support.

def shortrepr(obj):
    if obj is None:
        return repr(None)
    else:
        return obj.__shortrepr__()

# Mix-ins and abstract classes.

class Naming:

    "A mix-in providing naming conveniences."

    def full_name(self):
        if self.name is not None:
            return self.parent.full_name() + "." + self.name
        else:
            return self.parent.full_name()

no_attributes = {}

class Namespace:

    "A mix-in providing basic namespace functionality."

    def all_attributes(self):

        """
        Return all attributes accessible through a namespace, whether provided
        directly by the namespace or by other mechanisms.
        """

        return no_attributes

class Constant:

    "A superclass for all constant or context-free structures."

    pass

# Instances are special in that they need to be wrapped together with context in
# a running program, but they are not generally constant.

class Instance(Namespace):

    "A placeholder indicating the involvement of an instance."

    def __init__(self):
        self.parent = None

    def __repr__(self):
        return "<instance>"

    def __eq__(self, other):
        return other.__class__ is Instance

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return 0

    __shortrepr__ = __repr__

# Common instance construction.

common_instance = Instance()

def make_instance():
    return common_instance

class TypedInstance(Instance):

    "A placeholder indicating the involvement of an instance of a known type."

    def __init__(self, cls):
        Instance.__init__(self)
        self.cls = cls

    def __repr__(self):
        return "<instance of %s>" % shortrepr(self.cls)

    def __eq__(self, other):
        return other.__class__ is TypedInstance and other.cls is self.cls

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return 0

    __shortrepr__ = __repr__

# vim: tabstop=4 expandtab shiftwidth=4
