#!/usr/bin/env python

"""
Common classes.

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.stdcompiler import compiler
from micropython.basicdata import Constant, TypedInstance, Instance
from micropython.data import BaseAttr, Class
from micropython.errors import *
from os.path import split

try:
    set
except NameError:
    from sets import Set as set

# Visitors and activities related to node annotations.

class ASTVisitor:

    "A base class for visitors."

    def process_definitions(self, node):

        """
        Process and return all definitions beneath 'node', but not definitions
        within definitions.
        """

        definitions = []
        for n in node.getChildNodes():
            if isinstance(n, (compiler.ast.Class, compiler.ast.Function)):
                self.current_definition = n
                definitions.append(self.dispatch(n))
            else:
                definitions += self.process_definitions(n)
        return definitions

    def processing_definition(self, n):
        return self.current_definition is n

    # Visitor support methods.

    def default(self, node, *args):
        for n in node.getChildNodes():
            self.dispatch(n)

    def dispatch(self, node, *args):

        "Dispatch using 'node', annotating any raised exceptions."

        # Dispatch via a generic visit method.

        try:
            return node.visit(self, *args)

        # Annotate the exception in case of failure.

        except NodeProcessingError, exc:
            if exc.astnode is None:
                exc.astnode = node
                exc.unit_name = self.get_unit().full_name()
            raise

    def get_attributes(self, targets, attrname):

        "Return a list of attributes for 'targets' supporting 'attrname'."

        attributes = set()

        for target in targets:
            try:
                attributes.add(self.objtable.access(target.full_name(), attrname))
            except TableError:
                pass

        return attributes

    def get_attribute_and_value(self, obj):

        """
        Return (attribute, value) details for the given 'obj', where an
        attribute of None can be returned for constant objects, and where None
        can be returned as the result where no concrete details can be provided.
        """

        if isinstance(obj, (Constant, TypedInstance)):
            return None, obj

        if isinstance(obj, BaseAttr):
            return obj, obj.get_value()

        return None

    def get_type_for_attribute(self, attr, value):

        """
        For an attribute 'attr' and 'value', return the "parent" type of the
        attribute if possible. Otherwise, return None.
        """

        if attr:
            return attr.get_type()
        elif value and not isinstance(value, Instance) and value.parent:
            return value.parent
        else:
            return None

    def get_values_for_attribute(self, attr, value):

        """
        Return values defined for the given attribute 'attr' unless a specific
        'value' is provided, in which case a list containing that value is
        returned. Where no values are defined, return None in a list.
        """

        return value and [value] or attr and attr.get_values() or [None]

def get_module_name(node, module):

    """
    Using the given From 'node' and 'module' in which it is found, calculate
    any relative import information, returning a tuple containing a module
    to import along with any names to import based on the node's name
    information.

    Where the returned module is given as None, whole module imports should
    be performed for the returned modules using the returned names.
    """

    # Absolute import.

    if node.level == 0:
        return node.modname, node.names

    # Relative to an ancestor of this module.

    else:
        path = module.full_name().split(".")
        level = node.level

        # Relative imports treat package roots as submodules.

        if split(module.filename)[-1] == "__init__.py":
            level -= 1

        if level > len(path):
            raise InspectError("Relative import %r involves too many levels up from module %r" % (
                ("%s%s" % ("." * node.level, node.modname or "")), module.full_name()))

        basename = ".".join(path[:len(path)-level])

    # Name imports from a module.

    if node.modname:
        return "%s.%s" % (basename, node.modname), node.names

    # Relative whole module imports.

    else:
        return None, [("%s.%s" % (basename, name), alias or name) for name, alias in node.names]

def used_by_unit(node):

    """
    Return whether the definition made by a 'node' is actually employed by the
    program unit within which it is found.
    """

    return node.unit and node.unit.parent.has_key(node.unit.original_name)

# Useful data.

operator_functions = {

    # Fundamental operations.

    "in" : "in_",
    "not in" : "not_in",

    # Binary operations.

    "Add" : "add",
    "Bitand" : "and_",
    "Bitor" : "or_",
    "Bitxor" : "xor",
    "Div" : "div",
    "FloorDiv" : "floordiv",
    "LeftShift" : "lshift",
    "Mod" : "mod",
    "Mul" : "mul",
    "Power" : "pow",
    "RightShift" : "rshift",
    "Sub" : "sub",

    # Unary operations.

    "Invert" : "invert",
    "UnaryAdd" : "pos",
    "UnarySub" : "neg",

    # Augmented assignment.

    "+=" : "iadd",
    "-=" : "isub",
    "*=" : "imul",
    "/=" : "idiv",
    "//=" : "ifloordiv",
    "%=" : "imod",
    "**=" : "ipow",
    "<<=" : "ilshift",
    ">>=" : "irshift",
    "&=" : "iand",
    "^=" : "ixor",
    "|=" : "ior",

    # Comparisons.

    "==" : "eq",
    "!=" : "ne",
    "<" : "lt",
    "<=" : "le",
    ">=" : "ge",
    ">" : "gt",

    # Access and slicing.

    "AssSlice" : "setslice",    # Python 2.7
    "Slice" : "getslice",
    "AssSubscript" : "setitem", # Python 2.7
    "Subscript" : "getitem",
    }

# vim: tabstop=4 expandtab shiftwidth=4
