#!/usr/bin/env python

"""
Perform deductions on an inspected program.

Copyright (C) 2006, 2007, 2010, 2011, 2012, 2013, 2014 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.stdcompiler import compiler
from compiler.ast import *

from micropython.basicdata import Constant, TypedInstance
from micropython.common import ASTVisitor, used_by_unit
from micropython.data import *
from micropython.errors import *

try:
    set
except NameError:
    from sets import Set as set

# Source code classes.

class DeducedSource(ASTVisitor):

    "A module upon which deductions of code behaviour are made."

    def __init__(self, module, program):
        self.visitor = self
        self.module = module
        self.program = program
        self.objtable = program.get_object_table()
        self.units = []
        self.expr = None

    def get_unit(self):
        return self.units[-1]

    def get_module(self):
        return self.units[0]

    def deduce(self):

        "Process the module, making deductions."

        self.dispatch(self.module.astnode)

    def dispatch(self, node, *args):

        "NOTE: From compiler.visitor.ASTVisitor."

        try:
            return node.visit(self.visitor, *args)
        except AttributeError:
            # NOTE: Obligatory hack to find real attribute errors.
            if isinstance(node, self.implemented_nodes):
                raise
            return self.visitor.default(node, *args)

    implemented_nodes = (
        AssAttr, Assign, AssName, AssList, AssTuple, CallFunc, Function, Getattr,
        Add, Bitand, Bitor, Bitxor, Div, FloorDiv, Invert, LeftShift, Mod, Mul,
        Power, RightShift, Sub, UnaryAdd, UnarySub
        )

    # Deduction-related methods.

    def provides_constant_result(self, value):

        "Return whether 'value' provides a constant result."

        return isinstance(value, (Const, Constant))

    def provides_self_access(self, expr, unit):

        """
        Return whether the 'expr' in the given 'unit' provides a self-based
        attribute access.
        """

        attr_value = self.get_attribute_and_value(expr)

        if attr_value:
            target, value = attr_value

            return target and target.name == "self" and target.parent is unit and \
                unit.is_method()

        return False

    def possible_attributes_from_annotation(self, expr, attr, attrname):

        """
        Return (attribute, value) details provided by the 'expr' or 'attr'
        annotations on a node for an access involving 'attrname'.
        """

        attr_value = self.get_attribute_and_value(attr)

        if attr_value:
            return [attr_value]

        attrs = set()

        if expr:

            # Permitting multiple expression types if they provide the
            # attribute.

            if isinstance(expr, BaseAttr):
                exprs = expr.get_values()
            else:
                exprs = [expr]

            # For each expression value try and get a concrete attribute.

            for _expr in exprs:
                if isinstance(_expr, Instance):
                    continue

                found_attr = _expr.all_attributes().get(attrname)

                # Where an attribute can be obtained, record its
                # details.

                if found_attr:
                    attrs.add((found_attr, found_attr.get_value()))

        return attrs

    def possible_accessor_types_from_usage(self, node, defining_users=1):

        """
        Return a set of (target name, static) tuples from an investigation of
        attribute usage observations stored on the given 'node'.

        If 'defining_users' is set to a false value, attempt to get the type
        names specifically applicable to the node, rather than retrieving more
        general definition-based type observations.
        """

        target_names = set()

        if node._attrusers:

            # Visit each attribute user.

            for user in node._attrusers:

                # Since users such as branches may not provide type information,
                # attempt to find defining users.

                if defining_users:
                    for def_user in user._attrdefs or [user]:
                        for target_name, is_static in def_user._attrtypes.get(node._username, []):
                            target_names.add((target_name, is_static))
                else:
                    for target_name, is_static in user._attrspecifictypes.get(node._username, []):
                        target_names.add((target_name, is_static))

        return target_names

    def possible_accessors_from_usage(self, node, defining_users=1):

        """
        Return possible accessors from the usage recorded on the given 'node'.

        If 'defining_users' is set to a false value, attempt to get the type
        names specifically applicable to the node, rather than retrieving more
        general definition-based type observations.
        """

        target_names = self.possible_accessor_types_from_usage(node, defining_users)
        return self.get_targets_from_type_names(target_names)

    def possible_accessors_for_attribute(self, attrname):

        "Return possible accessors given the single 'attrname'."

        targets = set()

        for target_name in self.objtable.any_possible_objects([attrname]):
            targets.add(self.objtable.get_object(target_name))

        return targets

    def get_targets_from_type_names(self, target_names):

        """
        Given a collection of 'target_names' of the form (name, is_static),
        return a set of types for the 'name' part of each tuple.
        """

        targets = set()

        if target_names:
            for target_name, is_static in target_names:
                targets.add(self.objtable.get_object(target_name))

        return targets

    # Visitor methods.

    def _visitUnit(self, node):

        """
        Track entry into program units in order to support various attribute
        access operations.
        """

        self.units.append(node.unit)
        self.dispatch(node.node)
        self.units.pop()

    visitModule = _visitUnit

    def visitClass(self, node):

        "Optionally visit a class, depending on whether it is used."

        if not used_by_unit(node):
            return
        self._visitUnit(node)

    def visitFunction(self, node):

        "Optionally visit a function, depending on whether it is used."

        if not used_by_unit(node):
            return

        self.units.append(node.unit)
        self.dispatch(node.code)
        self.units.pop()

        node._guard_types = {}
        node._guards = {}

        self._visitParameters(node, node.unit.positional_names)

    def _visitParameters(self, node, parameters):
        for name in parameters:
            if isinstance(name, tuple):
                self._visitParameters(node, name)
            else:
                # NOTE: No _values usage here because it is mostly useless information, except perhaps for defaults.
                types = self.get_targets_from_type_names(node._attrtypes.get(name))
                self._visitGuardForNameDeduced(node, name, types)

    def _visitAttr(self, node):

        """
        Perform deductions on attribute accesses, adding annotations to the node
        that can be used by subsequent activities.
        """

        # Remember to permit deductions on the expression node. Here, we may
        # also obtain a concrete type associated with an instantiation.

        expr_type = self.dispatch(node.expr)
        if not node._expr or isinstance(node._expr, Instance):
            node._expr = expr_type

        self._annotateAttr(node, node._expr, node.attrname)

        # Make a summary annotation of the deductions.

        attributes = node._value_deduced and [self.get_attribute_and_value(node._value_deduced)] or \
                     node._attr_deduced and [self.get_attribute_and_value(node._attr_deduced)] or \
                     node._attrs_deduced or \
                     map(self.get_attribute_and_value, node._attrs_deduced_from_specific_usage or [])

        node._access_attrs = attributes

        # Return the _attr annotation for instantiation detection.

        return node._attr

    def _annotateAttr(self, node, target, attrname):

        """
        Annotate the access on the given 'node' using the 'target' and
        'attrname' information.
        """

        unit = self.get_unit()

        # The target, on which the access is performed, may influence the effect
        # on the context. We can only reliably assume that a literal constant is
        # an instance: all other "instances" may actually be classes in certain
        # cases.

        instance_target = isinstance(target, TypedInstance)
        typed_instance_attr = isinstance(target, BaseAttr) and isinstance(target.get_value(), TypedInstance)
        self_access = self.provides_self_access(target, unit)

        # Attempt to deduce attributes from explicit annotations.

        node._attrs_deduced = attrs = self.possible_attributes_from_annotation(target, node._attr, attrname)

        if len(attrs) == 1:
            for attr, value in attrs:

                # Constant values can be obtained directly.

                if self.provides_constant_result(value):
                    node._access_type = "constant"
                    node._value_deduced = value
                    return

                # Static attributes can be obtained via their parent.

                if attr.is_static_attribute():
                    node._access_type = "static"
                    node._attr_deduced = attr
                    node._set_context = instance_target and "set" or None
                    return

        # If a reliable target was originally specified, any usable attributes
        # should have been detected above, and any attributes deduced by other
        # means will not be compatible with the target. Thus, the nature of the
        # target is investigated: it must be an inspectable namespace or be an
        # attribute only providing such namespaces; otherwise, it is possible
        # that deduced attributes might be appropriate.

        if target and (
            isinstance(target, (Class, Module)) or
            isinstance(target, BaseAttr) and not [v for v in target.get_values() if not isinstance(v, (Class, Module))]
            ):
            node._access_type = "impossible"
            return

        # Attributes of self, which is by definition an instance, or typed
        # instances, which act somewhat like self in that their properties
        # should be known.

        if instance_target or typed_instance_attr or self_access:

            if instance_target:
                value = target
            elif typed_instance_attr:
                value = target.get_value()

            # Find the class of the instance.

            if instance_target or typed_instance_attr:
                if isinstance(value, Const):
                    cls = get_constant_class(value.get_class_name())
                else:
                    cls = value.cls
            else:
                cls = unit.parent

            # Find instance attributes.

            attr = cls.instance_attributes().get(attrname)

            # Where self is involved, descendants can also provide attributes.

            attrs = self_access and filter(None, [desc.instance_attributes().get(attrname) for desc in cls.descendants]) or []

            # A "leaf" class whose instances provide an attribute.

            if attr and not attrs:
                node._access_type = "instance"
                node._attr_deduced = attr
                return

            # A class where instances of subclasses may provide an attribute.

            elif attrs and self_access:
                if attr:
                    attrs.append(attr)

                node._attrs_deduced = [(a, a.get_value()) for a in attrs]

                # The instances may provide the attribute at the same position.

                positions = set([a.position for a in attrs])
                if len(positions) == 1:
                    for position in positions:
                        node._access_type = "positioned"
                        node._position_deduced = position
                        return

                # Otherwise, accessing the attributes is more work.

                node._access_type = "instance"
                return

            # Find class attributes.
            # The context will be overridden for compatible class attributes
            # only.

            attr = cls.all_class_attributes().get(attrname)

            if attr:

                # Constant attributes.

                if attr.is_strict_constant():
                    if self.provides_constant_result(attr.get_value()):
                        node._access_type = "constant"
                        node._value_deduced = attr.get_value()
                        node._set_context = "set"
                        return

                # Compatible class attributes.

                if attr.defined_within_hierarchy():
                    node._access_type = "static"
                    node._attr_deduced = attr
                    node._set_context = "set"
                    return

                # Incompatible class attributes.

                elif attr.defined_outside_hierarchy():
                    node._access_type = "static"
                    node._attr_deduced = attr
                    return

                # Unknown or mixed compatibility.

                node._access_type = "static"
                node._attr_deduced = attr
                node._set_context = "cond"
                return

        # Usage observations, both specific to this node's region of the program
        # and also applicable to the lifespan of the affected name.

        specific_targets = self.possible_accessors_from_usage(node, defining_users=0)
        targets = self.possible_accessors_from_usage(node, defining_users=1)

        # Record whether types were already deduced. If not, get types using
        # only this attribute.

        if not specific_targets or not targets:
            attribute_targets = self.possible_accessors_for_attribute(attrname)
            if not specific_targets:
                specific_targets = attribute_targets
            if not targets:
                targets = attribute_targets

        # Get the attributes from the deduced targets.

        node._attrs_deduced_from_specific_usage = self.get_attributes(specific_targets, attrname)
        node._attrs_deduced_from_usage = attrs = self.get_attributes(targets, attrname)

        # Generate optimisations where only a single attribute applies.

        if attrs and len(attrs) == 1:
            for attr in attrs:

                # Static attributes, but potentially non-static targets.

                if attr.is_static_attribute():

                    # Static attributes may be accompanied by a different context
                    # depending on the accessor.
                    # NOTE: Should determine whether the context is always replaced.

                    node._access_type = "static"
                    node._attr_deduced = attr
                    node._set_context = instance_target and "set" or "cond"
                    return

                # Non-static attributes.

                node._access_type = "instance"
                node._attr_deduced = attr
                return

        # Test for compatible attribute positioning.

        elif attrs:
            positions = set([(attr.is_static_attribute(), attr.position) for attr in attrs])

            # Permit a position-based access only on non-static attributes since
            # access to static attributes may happen via instances and thus not
            # be relative to the accessor but to its parent.

            if len(positions) == 1:
                for position in positions:
                    if not position[0]:
                        node._access_type = "positioned"
                        node._position_deduced = position[0]
                        return

        # With no usable deductions, generate a table-based access.

        node._access_type = "unknown"
        node._set_context = "cond"

    visitAssAttr = visitGetattr = _visitAttr

    def visitAssign(self, node):
        self.expr = self.dispatch(node.expr)
        for n in node.nodes:
            self.dispatch(n)

    def visitAssList(self, node):
        expr = self.expr
        self.expr = make_instance()
        for n in node.nodes:
            self.dispatch(n)
        self.expr = expr

    visitAssTuple = visitAssList

    def visitAssName(self, node):
        if self.expr:
            if isinstance(self.expr, BaseAttr):
                expr = self.expr.get_value()
            elif isinstance(self.expr, TypedInstance):
                expr = self.expr
            else:
                self._visitGuard(node)
                return
        else:
            self._visitGuard(node)
            return

        attr = node._values and node._values.get(node.name) or None

        # Need to replace any uncertain value with a concrete value.

        if attr:
            if isinstance(attr, BaseAttr):
                value = attr.get_value()
            else:
                value = attr

            if value and isinstance(value, Instance) and not isinstance(value, TypedInstance):
                node._values[node.name] = expr or make_instance()

        self._visitGuard(node)

    def _visitGuard(self, node):
        node._guard_types = {}
        node._guards = {}
        self._visitGuardForName(node, node.name)

    def _visitGuardForName(self, node, name):

        "Make an annotation stating the acceptable types for a name."

        # Need to check any concrete value against deductions.
        # Where no local observations exist, no guard can usefully be asserted.

        if not node._attrtypes and not node._values:
            return

        types = self.get_targets_from_type_names(node._attrtypes.get(name))
        value = node._values.get(name)

        if isinstance(value, BaseAttr):
            values = value.get_values()
        else:
            values = [value]

        # Obtain values with identifiable types.

        typed_values = set()

        for v in values:
            if isinstance(v, Instance):
                if not isinstance(v, TypedInstance):
                    return
                typed_values.add(v.cls)
            else:
                typed_values.add(v)

        # Where a concrete type conflicts with deductions, the usage of
        # attributes cannot be supported and is thus impossible.

        if typed_values:
            if types and not typed_values.intersection(types):
                node._guard_types[name] = "impossible"
            else:
                node._guard_types[name] = len(typed_values) == 1 and "single" or "multiple"
                node._guards[name] = typed_values

        # Where no concrete type is known, usage must be checked.

        elif types:
            self._visitGuardForNameDeduced(node, name, types)

        # Where no deductions are made, no guards are needed.

    def _visitGuardForNameDeduced(self, node, name, types):
        if not types:
            return
        if len(types) == 1:
            node._guard_types[name] = "single"
        else:
            node._guard_types[name] = "multiple"
        node._guards[name] = types

    def visitCallFunc(self, node):

        "Identify any concrete types involved with instantiation."

        for n in node.getChildNodes()[1:]: # ignore node.node
            self.dispatch(n)

        # Determine whether the target of the invocation refers to a class.

        attr = self.dispatch(node.node)

        if attr:
            if not isinstance(attr, (Instance, UnresolvedName)):
                value = attr.get_value()
                if value and isinstance(value, Class):
                    return TypedInstance(value)

    def _visitOperator(self, node):

        "Annotate operators with function information."

        self._annotateAttr(node, node._module, node._attr.name)

    visitAdd = \
    visitBitand = \
    visitBitor = \
    visitBitxor = \
    visitDiv = \
    visitFloorDiv = \
    visitInvert = \
    visitLeftShift = \
    visitMod = \
    visitMul = \
    visitPower = \
    visitRightShift = \
    visitSub = \
    visitUnaryAdd = \
    visitUnarySub = _visitOperator

# Convenience functions.

def deduce(program):
    for module in program.get_importer().get_modules():
        DeducedSource(module, program).deduce()

# vim: tabstop=4 expandtab shiftwidth=4
